/* {=================================================================
 *
 * fmindex.c
 * FM index, a compressed full-text index based on the BWT, in PostBio
 * This index is proposed by Ferragina and Manzini; more details at
 * [ http://roquefort.di.unipi.it/~ferrax/fmindexV2/index.html ]
 *
 * Author: Luis Carvalho <lexcarvalho at gmail dot com>
 * Please check copyright notice at the bottom of postbio.sql
 *
 * ==================================================================} */

#include <postgres.h>
#include <fmgr.h>
#include <funcapi.h>
#include <miscadmin.h>
#include <access/heapam.h>
#include <utils/builtins.h>
#include <libpq/pqformat.h>

#include "fmindexV2/interface.h"

#if PG_VERSION_NUM < 80300
#define SET_VARSIZE(ptr, len) VARATT_SIZEP(ptr) = len
#endif

typedef struct varlena FMindex;

#define COMPRESS(F) ((uchar *) VARDATA(F)) /* compressed index */
#define BUILDOPT "-a 2" /* internal copy of text */
#define ASSERT(stmt) { int e = stmt; if (e) elog(ERROR, "%s", error_index(e)); }

typedef struct {
  Uint c; /* current occurrence */
  Uint n; /* number of occurrences */
  Uint *occ;
} fm_locate_info;


static Datum buildfmindex (uchar *text, Uint l) {
  FMindex *F;
  void *index;
  int32 size; /* varlena size */
  Uint indexsize;
  /* build index */
  ASSERT(build_index(text, l, BUILDOPT, &index));
  index_size(index, &indexsize);
  /* alloc Datum */
  size = VARHDRSZ + indexsize;
  F = (FMindex *) palloc(size);
  SET_VARSIZE(F, size);
  /* copy index */
  ASSERT(save_index_mem(index, COMPRESS(F)));
  free_index(index);
  PG_RETURN_POINTER(F);
}

static uchar *extracttext (FMindex *F, Uint start, Uint len, Uint *l) {
  void *index;
  Uint indexsize = VARSIZE(F) - VARHDRSZ;
  uchar *text;
  start--; /* zero based */
  ASSERT(load_index_mem(&index, COMPRESS(F), indexsize));
  ASSERT(extract(index, start, start + len - 1, &text, l));
  free_index(index);
  return text;
}

static uchar *extractfulltext (FMindex *F, Uint *l) {
  void *index;
  Uint indexsize = VARSIZE(F) - VARHDRSZ;
  Uint l0; /* full text size */
  uchar *text;
  ASSERT(load_index_mem(&index, COMPRESS(F), indexsize));
  ASSERT(get_length(index, &l0));
  ASSERT(extract(index, 0, l0 - 1, &text, l));
  free_index(index);
  return text;
}



/* {=====================================================================
 *    Prototypes
 * ======================================================================} */

Datum fmindex_in (PG_FUNCTION_ARGS);
Datum fmindex_out (PG_FUNCTION_ARGS);
Datum fmindex_recv (PG_FUNCTION_ARGS);
Datum fmindex_bytea (PG_FUNCTION_ARGS);
Datum fmindex_send (PG_FUNCTION_ARGS);
Datum fmindex_build (PG_FUNCTION_ARGS);
Datum fmindex_text (PG_FUNCTION_ARGS);
Datum fmindex_extract (PG_FUNCTION_ARGS);
Datum fmindex_length (PG_FUNCTION_ARGS);
Datum fmindex_count (PG_FUNCTION_ARGS);
Datum fmindex_locate (PG_FUNCTION_ARGS);

/* {=====================================================================
 *    Functions
 * ======================================================================} */

PG_FUNCTION_INFO_V1(fmindex_in);
Datum fmindex_in (PG_FUNCTION_ARGS) {
  uchar *text = (uchar *) PG_GETARG_CSTRING(0);
  Uint l = (Uint) strlen((const char *) text);
  return buildfmindex(text, l);
}

PG_FUNCTION_INFO_V1(fmindex_out);
Datum fmindex_out (PG_FUNCTION_ARGS) {
  FMindex *F = (FMindex *) PG_GETARG_VARLENA_P(0);
  Uint l; /* extracted text length */
  uchar *text = extractfulltext(F, &l);
  uchar *result = (uchar *) palloc(l + 1);
  memcpy(result, text, l);
  free(text);
  result[l] = '\0';
  PG_RETURN_CSTRING(result);
}

PG_FUNCTION_INFO_V1(fmindex_recv);
Datum fmindex_recv (PG_FUNCTION_ARGS) {
  StringInfo buf = (StringInfo) PG_GETARG_POINTER(0);
  FMindex *F;
  int32 size = buf->len - buf->cursor;
  F = (FMindex *) palloc(VARHDRSZ + size);
  SET_VARSIZE(F, VARHDRSZ + size);
  pq_copymsgbytes(buf, VARDATA(F), size);
  PG_RETURN_POINTER(F);
}

PG_FUNCTION_INFO_V1(fmindex_bytea);
Datum fmindex_bytea (PG_FUNCTION_ARGS) {
  FMindex *F = (FMindex *) PG_GETARG_BYTEA_P_COPY(0);
  PG_RETURN_POINTER(F);
}

PG_FUNCTION_INFO_V1(fmindex_send);
Datum fmindex_send (PG_FUNCTION_ARGS) {
  FMindex *F = (FMindex *) PG_GETARG_VARLENA_P(0);
  StringInfoData buf;
  pq_begintypsend(&buf);
  pq_sendbytes(&buf, VARDATA(F), VARSIZE(F) - VARHDRSZ);
  PG_RETURN_BYTEA_P(pq_endtypsend(&buf));
}

PG_FUNCTION_INFO_V1(fmindex_build);
Datum fmindex_build (PG_FUNCTION_ARGS) {
  text *t = PG_GETARG_TEXT_P(0);
  uchar *text = (uchar *) VARDATA(t);
  Uint l = VARSIZE(t) - VARHDRSZ;
  return buildfmindex(text, l);
}

PG_FUNCTION_INFO_V1(fmindex_text);
Datum fmindex_text (PG_FUNCTION_ARGS) {
  FMindex *F = (FMindex *) PG_GETARG_VARLENA_P(0);
  Uint l; /* extracted text length */
  uchar *t = extractfulltext(F, &l);
  text *result = (text *) palloc(VARHDRSZ + l);
  SET_VARSIZE(result, VARHDRSZ + l);
  memcpy(VARDATA(result), t, l);
  free(t);
  PG_RETURN_TEXT_P(result);
}


PG_FUNCTION_INFO_V1(fmindex_extract);
Datum fmindex_extract (PG_FUNCTION_ARGS) {
  FMindex *F = (FMindex *) PG_GETARG_VARLENA_P(0);
  Uint start = PG_GETARG_INT32(1);
  Uint len = PG_GETARG_INT32(2);
  Uint l; /* extracted text length */
  uchar *t = extracttext(F, start, len, &l);
  text *result = (text *) palloc(VARHDRSZ + l);
  SET_VARSIZE(result, VARHDRSZ + l);
  memcpy(VARDATA(result), t, l);
  free(t);
  PG_RETURN_TEXT_P(result);
}

PG_FUNCTION_INFO_V1(fmindex_length);
Datum fmindex_length (PG_FUNCTION_ARGS) {
  FMindex *F = (FMindex *) PG_GETARG_VARLENA_P(0);
  Uint indexsize = VARSIZE(F) - VARHDRSZ;
  void *index;
  Uint l;
  ASSERT(load_index_mem(&index, COMPRESS(F), indexsize));
  ASSERT(get_length(index, &l));
  free_index(index);
  PG_RETURN_INT32(l);
}


PG_FUNCTION_INFO_V1(fmindex_count);
Datum fmindex_count (PG_FUNCTION_ARGS) {
  FMindex *F = (FMindex *) PG_GETARG_VARLENA_P(0);
  text *p = PG_GETARG_TEXT_P(1);
  uchar *pattern = (uchar *) VARDATA(p);
  Uint patternlen = VARSIZE(p) - VARHDRSZ;
  Uint indexsize = VARSIZE(F) - VARHDRSZ;
  Uint c; /* number of occurrences */
  void *index;
  ASSERT(load_index_mem(&index, COMPRESS(F), indexsize));
  ASSERT(count(index, pattern, patternlen, &c));
  free_index(index);
  PG_RETURN_INT32(c);
}

PG_FUNCTION_INFO_V1(fmindex_locate);
Datum fmindex_locate (PG_FUNCTION_ARGS) {
  FuncCallContext *funcctx;
  fm_locate_info *fminfo;
  if (SRF_IS_FIRSTCALL()) {
    MemoryContext oldcontext;
    FMindex *F = (FMindex *) PG_GETARG_VARLENA_P(0);
    text *p = PG_GETARG_TEXT_P(1);
    uchar *pattern = (uchar *) VARDATA(p);
    Uint patternlen = VARSIZE(p) - VARHDRSZ;
    Uint indexsize = VARSIZE(F) - VARHDRSZ;
    void *index;
    /* setup funcctx */
    funcctx = SRF_FIRSTCALL_INIT();
    oldcontext = MemoryContextSwitchTo(funcctx->multi_call_memory_ctx);
    funcctx->user_fctx = palloc(sizeof(fm_locate_info));
    fminfo = (fm_locate_info *) funcctx->user_fctx;
    fminfo->c = 0;
    /* locate */
    ASSERT(load_index_mem(&index, COMPRESS(F), indexsize));
    ASSERT(locate(index, pattern, patternlen, &fminfo->occ, &fminfo->n));
    free_index(index);
    MemoryContextSwitchTo(oldcontext);
  }
  funcctx = SRF_PERCALL_SETUP();
  fminfo = (fm_locate_info *) funcctx->user_fctx;
  if (fminfo->c < fminfo->n)
    SRF_RETURN_NEXT(funcctx, Int32GetDatum(fminfo->occ[fminfo->c++] + 1));
  else {
    if (fminfo->n > 0) free(fminfo->occ);
    SRF_RETURN_DONE(funcctx);
  }
}

